#ifndef BOOZ_INIT_TRIG_H
#define BOOZ_INIT_TRIG_H

#include "vector_matrix.hpp"

namespace booz_xform {  

  void init_trig(Vector&, Vector&, Matrix&, Matrix&, Matrix&, Matrix&, int, int, int);

}

#endif

